# The MIT License (MIT)
#
# Copyright (C) 2025 Fabrício Barros Cabral
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Config: load and save config parameters from a config file.

Returns:
    Config: a Config

"""

import json
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any


class Config(ABC):
    """Interface that represents a Config.

    Returns:
        Config: a Config

    """

    @abstractmethod
    def load(self) -> dict[str, Any]:
        """Load the config params.

        Returns:
            dict[str, Any]: a dictionary with code and pages

        """
        pass

    @abstractmethod
    def save(self, data: dict[str, Any]) -> None:
        """Save the config params in a file.

        Args:
            data: a dictionary with code and pages

        """
        pass

    @abstractmethod
    def __str__(self) -> str:
        """Convert the object in a string.

        Returns:
            str: the config file content as a string

        """
        pass


class FileConfig(Config):
    """A class that represents a Config.

    Returns:
        Config: a Config

    """

    def __init__(self, filename: str = ".zapspiderrc") -> None:
        """Initialize a Config.

        Args:
            filename (str, optional): The config filename. Defaults to
            '.zapspcodeerrc'.

        """
        self.__filename = filename

    def load(self) -> dict[str, Any]:
        """Load the config params.

        Returns:
            dict[str, Any]: a dictionary with code and pages

        """
        return json.loads(Path(self.__filename).read_text(encoding="utf-8"))

    def save(self, data: dict[str, Any]) -> None:
        """Save the config params in a file.

        Args:
            data: a dictionary with code and pages

        """
        Path(self.__filename).write_text(
            json.dumps(data, ensure_ascii=False), encoding="utf-8"
        )

    def __str__(self) -> str:
        """Convert the object in a string.

        Returns:
            str: the config file content as a string

        """
        return self.__filename


class FakeConfig(Config):
    """A class that represents a Config.

    Returns:
        Config: a Config

    """

    def __init__(self, data: dict[str, Any]) -> None:
        self.__data = data

    def load(self) -> dict[str, Any]:
        """Load the config params.

        Returns:
            dict[str, Any]: a dictionary with code and pages

        """
        return self.__data

    def save(self, data: dict[str, Any]) -> None:
        """Save the config params in a file.

        Args:
            data: a dictionary with code and pages

        """
        self.__data = data

    def __str__(self) -> str:
        """Convert the object in a string.

        Returns:
            str: the config file content as a string

        """
        return str(self.__data)



# The MIT License (MIT)
#
# Copyright (C) 2025 Fabrício Barros Cabral
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import re

from fake_https_server.request import ContentGet, Fail, FileContentGet
from fake_https_server.server import Daemon, FakeHttpsServer

import arana.browser
import arana.page
from arana.browser import Chromium
from arana.console import FakeConsole
from arana.rndm import RandomWait


def test_url() -> None:
    url = "https://www.example.com"
    browser = Chromium()
    browser.open()
    page = browser.page(url)
    assert page.url() == url
    browser.close()


def test_not_open() -> None:
    browser = Chromium()
    browser.open()
    page = browser.page("https://www.example.com")
    page.close()
    browser.close()


def test_scroll() -> None:
    success_ok = 200
    server = Daemon(
        FakeHttpsServer(
            FileContentGet(
                "tests/contents/galpoes-depositos-armazens-alugar-sp.html"
            )
        )
    )
    browser = Chromium()
    browser.open()
    server.start()
    url = (
        f"https://localhost:{server.port()}/aluguel/galpao-deposito-armazem/sp/"
    )
    page = browser.page(url)
    response = page.open()
    done = page.scroll(RandomWait(0, 0))
    assert response.status() == success_ok
    assert done
    page.close()
    browser.close()
    server.stop()


def test_logged() -> None:
    server = Daemon(
        FakeHttpsServer(
            FileContentGet("tests/contents/armazem-arpoador-sp.html"),
        )
    )
    console = FakeConsole()
    browser = Chromium()
    browser.open()
    server.start()
    port = server.port()
    url = (
        f"https://localhost:{port}/imovel/aluguel-galpao-deposito"
        f"-armazem-com-cozinha-jardim-arpoador-zona-oeste-zona-oeste"
        f"-sao-paulo-sp-250m2-id-2743779110/"
    )
    page = arana.page.Logged(browser.page(url), console)
    page.open()
    page.close()
    browser.close()
    server.stop()
    assert re.search(
        r"\[[0-9]{2}:[0-9]{2}:[0-9]{2}\.[0-9]{6}] Opening "
        f"'https://localhost:{port}/imovel/aluguel-galpao-deposito-armazem-com-"
        r"cozinha-jardim-arpoador-zona-oeste-zona-oeste-sao-paulo-sp-250m2-id-"
        r"2743779110/'... done.\n"
        r"\[[0-9]{2}:[0-9]{2}:[0-9]{2}\.[0-9]{6}] Closing "
        f"'https://localhost:{port}/imovel/aluguel-galpao-deposito-armazem-com-"
        r"cozinha-jardim-arpoador-zona-oeste-zona-oeste-sao-paulo-sp-250m2-id-"
        r"2743779110/'... done.\n",
        console.stderr(),
    )


def test_retry() -> None:
    success_ok = 200
    retries = 1
    msg = "Retry Page Test!"
    server = Daemon(
        FakeHttpsServer(
            Fail(
                ContentGet(msg),
                retries,
            )
        )
    )
    browser = Chromium()
    browser.open()
    server.start()
    url = f"https://localhost:{server.port()}/index.html"
    page = arana.page.Retry(browser.page(url))
    response = page.open()
    assert response.status() == success_ok
    assert (
        response.html().content()
        == f"<html><head></head><body>{msg}</body></html>"
    )
    page.close()
    browser.close()
    server.stop()


def test_retry_logged() -> None:
    success_ok = 200
    retries = 1
    msg = "Retry Page Test!"
    server = Daemon(
        FakeHttpsServer(
            Fail(
                ContentGet(msg),
                retries,
            )
        )
    )
    browser = arana.browser.Logged(Chromium())
    browser.open()
    server.start()
    url = f"https://localhost:{server.port()}/index.html"
    page = arana.page.Retry(browser.page(url))
    response = page.open()
    assert response.status() == success_ok
    assert (
        response.html().content()
        == f"<html><head></head><body>{msg}</body></html>"
    )
    page.close()
    browser.close()
    server.stop()



# The MIT License (MIT)
#
# Copyright (C) 2025 Fabrício Barros Cabral
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from arana.rndm import RandomInt


def test_value() -> None:
    start = 0
    end = 10
    random_int = RandomInt()
    value = random_int.value(start, end)
    assert value >= start
    assert value <= end


def test_values() -> None:
    random_int = RandomInt()
    values = random_int.values(1, 10)
    assert sorted(values) == sorted([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])


def test_values_length_1() -> None:
    random_int = RandomInt()
    values = random_int.values(1, 1)
    assert len(values) == 1


def test_values_length_10() -> None:
    size = 10
    random_int = RandomInt()
    values = random_int.values(1, 10)
    assert len(values) == size



# The MIT License (MIT)
#
# Copyright (C) 2025 Fabrício Barros Cabral
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from abc import ABC, abstractmethod
from datetime import datetime, timedelta, timezone
from typing import Any

from arana.browser import Browser
from arana.console import Console, StdConsole
from arana.response import Response
from arana.rndm import RandomWait


class Content(ABC):
    @abstractmethod
    def url(self) -> str:
        pass

    @abstractmethod
    def refine(self, response: Response) -> dict[str, Any]:
        pass


class ContentText(Content):
    def __init__(self, url: str) -> None:
        self.__url = url

    def url(self) -> str:
        return self.__url

    def refine(self, response: Response) -> dict[str, Any]:
        html = response.html()
        text = html.evaluate("document.body.innerText")
        return {"html": html, "text": text}


class Logged(Content):
    def __init__(
        self, content: Content, console: Console = StdConsole()
    ) -> None:
        self.__origin = content
        self.__console = console

    def url(self) -> str:
        return self.__origin.url()

    def refine(self, response: Response) -> dict[str, Any]:
        zone = timezone(timedelta(hours=-3))
        timestamp = datetime.now(zone).strftime("%H:%M:%S.%f")
        self.__console.log(f"[{timestamp}] Refining '{self.url()}'... ")
        data = self.__origin.refine(response)
        self.__console.logln("done.")
        return data


class Retry(Content):
    def __init__(
        self,
        content: Content,
        browser: Browser,
        random_wait: RandomWait = RandomWait(2, 3),
        max_retries: int = 10,
    ) -> None:
        self.__origin = content
        self.__browser = browser
        self.__random_wait = random_wait
        self.__max_retries = max_retries

    def url(self) -> str:
        return self.__origin.url()

    def refine(self, response: Response) -> dict[str, Any]:
        data = None
        try:
            data = self.__origin.refine(response)
        except Exception:
            retries = 1
            while data is None and retries <= self.__max_retries:
                self.__random_wait.run()
                try:
                    url = self.__origin.url()
                    page = self.__browser.page(url)
                    resp = page.open()
                    data = self.__origin.refine(resp)
                    page.close()
                except Exception:
                    retries += 1
                    continue
                break
        return data



# The MIT License (MIT)
#
# Copyright (C) 2025 Fabrício Barros Cabral
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from arana.config import FakeConfig


def test_load() -> None:
    data = {"code": 10, "pages": [5, 3, 7, 9, 2, 56, 8, 19]}
    config = FakeConfig(data)
    params = config.load()
    assert params == data


def test_save() -> None:
    config = FakeConfig({})
    config.save({"code": 10, "pages": [5, 3, 7, 9, 2, 56, 8, 19]})
    assert str(config) == "{'code': 10, 'pages': [5, 3, 7, 9, 2, 56, 8, 19]}"



# The MIT License (MIT)
#
# Copyright (C) 2025 Fabrício Barros Cabral
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from abc import ABC, abstractmethod
from re import Pattern
from typing import Any

from playwright.sync_api import Page as Pwpg

from arana.element import Element, PwElement
from arana.frame import Frame, PwFrame


class Html(ABC):
    @abstractmethod
    def element(
        self, selector: str, *, has_text: str | Pattern[str] | None = None
    ) -> Element:
        pass

    @abstractmethod
    def by_role(
        self, role: str, *, name: str | Pattern[str] | None = None
    ) -> Element:
        pass

    @abstractmethod
    def by_test_id(self, test_id: str) -> Element:
        pass

    @abstractmethod
    def content(self) -> str:
        pass

    @abstractmethod
    def evaluate(self, expression: str) -> Any:
        pass

    @abstractmethod
    def frame(self, selector: str) -> Frame:
        pass


class PwHtml(Html):
    def __init__(self, pwpg: Pwpg) -> None:
        self.__pwpg = pwpg

    def element(
        self, selector: str, *, has_text: str | Pattern[str] | None = None
    ) -> Element:
        return PwElement(self.__pwpg.locator(selector, has_text=has_text))

    def by_role(
        self, role: str, *, name: str | Pattern[str] | None = None
    ) -> Element:
        return PwElement(self.__pwpg.get_by_role(role=role, name=name))

    def by_test_id(self, test_id: str) -> Element:
        return PwElement(self.__pwpg.get_by_test_id(test_id))

    def content(self) -> str:
        return self.__pwpg.content()

    def evaluate(self, expression: str) -> Any:
        return self.__pwpg.evaluate(expression)

    def frame(self, selector: str) -> Frame:
        return PwFrame(self.__pwpg.locator(selector).frame_locator())



# The MIT License (MIT)
#
# Copyright (C) 2025 Fabrício Barros Cabral
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Console: module to operate a console."""

import sys
from abc import ABC, abstractmethod


class Console(ABC):
    @abstractmethod
    def print(self, message: str) -> None:
        pass

    @abstractmethod
    def println(self, message: str) -> None:
        pass

    @abstractmethod
    def log(self, message: str) -> None:
        pass

    @abstractmethod
    def logln(self, message: str) -> None:
        pass

    @abstractmethod
    def stdout(self) -> str:
        pass

    @abstractmethod
    def stderr(self) -> str:
        pass


class StdConsole(Console):
    """Class to operate a console (stdout / stderr)."""

    def print(self, message: str) -> None:
        """Print a message flushed and without end in stdout.

        Args:
            message (str): a message

        """
        print(message, end="", flush=True)

    def println(self, message: str) -> None:
        """Print a message flushed in stdout.

        Args:
            message (str): a message

        """
        print(message, flush=True)

    def log(self, message: str) -> None:
        """Print a message flushed and without end in stderr.

        Args:
            message (str): a message

        """
        print(message, end="", flush=True, file=sys.stderr)

    def logln(self, message: str) -> None:
        """Print a message flushed and without end in stdout.

        Args:
            message (str): a message

        """
        print(message, flush=True, file=sys.stderr)

    def stdout() -> str:
        return ""

    def stderr() -> str:
        return ""


class FakeConsole(Console):
    def __init__(self) -> None:
        self.__stdout = []
        self.__stderr = []

    def print(self, message: str) -> None:
        self.__stdout.append(message)

    def println(self, message: str) -> None:
        self.__stdout.append(f"{message}\n")

    def log(self, message: str) -> None:
        self.__stderr.append(message)

    def logln(self, message: str) -> None:
        self.__stderr.append(f"{message}\n")

    def stdout(self) -> str:
        return "".join(self.__stdout)

    def stderr(self) -> str:
        return "".join(self.__stderr)



# The MIT License (MIT)
#
# Copyright (C) 2025 Fabrício Barros Cabral
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from abc import ABC, abstractmethod
from datetime import datetime, timedelta, timezone

from browserforge.fingerprints import FingerprintGenerator
from browserforge.injectors.playwright import NewContext
from playwright.sync_api import Browser as Rocket

from arana.console import Console, StdConsole
from arana.html import PwHtml
from arana.response import PwResponse, Response
from arana.rndm import RandomInt, RandomWait


class Page(ABC):
    @abstractmethod
    def url(self) -> str:
        pass

    @abstractmethod
    def open(self) -> Response:
        pass

    @abstractmethod
    def pause(self) -> None:
        pass

    @abstractmethod
    def close(self) -> None:
        pass

    @abstractmethod
    def reload(self) -> Response:
        pass

    @abstractmethod
    def scroll(self, random_wait: RandomWait) -> bool:
        pass


class PwPage(Page):
    def __init__(self, rocket: Rocket, url: str) -> None:
        self.__fingerprints = FingerprintGenerator()
        self.__session = NewContext(
            rocket,
            fingerprint=self.__fingerprints.generate(
                browser=("chrome", "firefox", "safari", "edge"),
                os=("windows", "macos", "linux"),
            ),
            ignore_https_errors=True,
        )
        self.__pwpg = self.__session.new_page()
        # I don't know why need it to allow close() without do a goto() before
        self.__pwpg.emulate_media(color_scheme="dark")
        self.__url = url

    def url(self) -> str:
        return self.__url

    def open(self) -> Response:
        self.__pwpg.wait_for_load_state()
        resp = self.__pwpg.goto(self.__url)
        return PwResponse(resp.status, PwHtml(self.__pwpg), self.__url)

    def pause(self) -> None:
        self.__pwpg.wait_for_event("close", timeout=0)

    def close(self) -> None:
        self.__pwpg.close()
        self.__session.close()

    def reload(self) -> Response:
        resp = self.__pwpg.reload()
        return PwResponse(resp.status, PwHtml(self.__pwpg), self.__url)

    def scroll(self, random_wait: RandomWait) -> bool:
        random_int = RandomInt()
        scroll_height = self.__pwpg.evaluate(
            "window.document.documentElement.scrollHeight"
        )
        pixels = 0
        prev_scroll_height = 0
        while prev_scroll_height < scroll_height:
            while pixels <= scroll_height:
                pixels += random_int.value(200, 600)
                self.__pwpg.evaluate(
                    f"() => window.scrollTo(0, {pixels}, {{ behavior: 'smooth' }})"
                )
                random_wait.run()
            prev_scroll_height = scroll_height
            scroll_height = self.__pwpg.evaluate(
                "window.document.documentElement.scrollHeight"
            )
        return True


class Logged(Page):
    def __init__(self, page: Page, console: Console = StdConsole()) -> None:
        self.__origin = page
        self.__console = console

    def url(self) -> str:
        return self.__origin.url()

    def open(self) -> Response:
        zone = timezone(timedelta(hours=-3))
        timestamp = datetime.now(zone).strftime("%H:%M:%S.%f")
        self.__console.log(f"[{timestamp}] Opening '{self.url()}'... ")
        response = self.__origin.open()
        self.__console.logln("done.")
        return response

    def pause(self) -> None:
        zone = timezone(timedelta(hours=-3))
        timestamp = datetime.now(zone).strftime("%H:%M:%S.%f")
        self.__console.logln(f"[{timestamp}] Page paused!")
        self.__origin.pause()

    def close(self) -> None:
        zone = timezone(timedelta(hours=-3))
        timestamp = datetime.now(zone).strftime("%H:%M:%S.%f")
        self.__console.log(f"[{timestamp}] Closing '{self.url()}'... ")
        self.__origin.close()
        self.__console.logln("done.")

    def reload(self) -> Response:
        zone = timezone(timedelta(hours=-3))
        timestamp = datetime.now(zone).strftime("%H:%M:%S.%f")
        self.__console.log(f"[{timestamp}] Reloading '{self.url()}'... ")
        response = self.__origin.reload()
        self.__console.logln("done.")
        return response

    def scroll(self, random_wait: RandomWait) -> bool:
        zone = timezone(timedelta(hours=-3))
        timestamp = datetime.now(zone).strftime("%H:%M:%S.%f")
        self.__console.log(f"[{timestamp}] Scrolling '{self.url()}'... ")
        state = self.__origin.scroll(random_wait)
        self.__console.logln("done.")
        return state


class Retry(Page):
    def __init__(
        self,
        page: Page,
        random_wait: RandomWait = RandomWait(2, 3),
        max_retries: int = 10,
    ) -> None:
        self.__origin = page
        self.__random_wait = random_wait
        self.__max_retries = max_retries

    def url(self) -> str:
        return self.__origin.url()

    def open(self) -> Response:
        try:
            response = self.__origin.open()
        except Exception:
            response = None
            retries = 1
            while response is None and retries <= self.__max_retries:
                self.__random_wait.run()
                try:
                    response = self.__origin.reload()
                except Exception:
                    retries += 1
                    continue
                break
        return response

    def pause(self) -> None:
        self.__origin.pause()

    def close(self) -> None:
        self.__origin.close()

    def reload(self) -> Response:
        return self.__origin.reload()

    def scroll(self, random_wait: RandomWait) -> bool:
        return self.__origin.scroll(random_wait)
